"use strict";
import * as fs from "fs";
import cache, { ICacheInsertData } from "../src/lib/cache";
import * as knexfile_cachedb from "../knexfile_cachedb";

const db = cache.db;

beforeAll(async () => {
  await db.migrate.latest();
});

afterAll(async () => {
  await db.destroy();
  try {
    fs.unlinkSync(knexfile_cachedb.connection.filename);
  } catch (e) {
    console.log(e);
  }
});

beforeEach(async () => {
  await db("cache").delete();
});

const universe_name = "testuniv",
  project_name = "testproj";

const ckey = "cacheKey1";
const validData: ICacheInsertData = {
  project_name,
  universe_name,
  ckey,
  ctype: "git_show",
  cvalue: "testValue",
  metadata: JSON.stringify({ foo: "bar" }),
};

// missing ckey
const invalidData = {
  project_name,
  universe_name,
  ctype: "git_show",
  cvalue: "testValue",
  metadata: JSON.stringify({ foo: "bar" }),
};

describe("cache.set()", () => {
  test("returns id & inserts correct values", async () => {
    let rez = await cache.set(validData);
    expect(Array.isArray(rez)).toBe(true);
    expect(await db("cache").first("*").where(validData)).toMatchObject(
      validData
    );
  });
  test("throws when passed invalid data", async () => {
    try {
      await cache.set(invalidData as ICacheInsertData);
    } catch (e) {
      return true;
    }
    return false;
  });
  test("overwrites an existing value if found", async () => {
    const newValue = `${validData.cvalue}_overwrite`;
    await cache.set(validData);
    await cache.set({ ...validData, cvalue: newValue });
    let rez = await db("cache").where({ universe_name, project_name, ckey });
    expect(rez.length).toEqual(1);
    expect(rez).toMatchObject([{ cvalue: newValue }]);
  });
});

describe("cache.get()", () => {
  test("returns correct data", async () => {
    await cache.set(validData);
    expect(
      await cache.get({ universe_name, project_name, ckey })
    ).toMatchObject(validData);
  });
  test("returns null when it can not find data", async () => {
    expect(
      await cache.get({ universe_name, project_name, ckey: "wrong" })
    ).toBeNull();
  });
});

describe("cache.getAll()", () => {
  test("returns many", async () => {
    await cache.set(validData);
    await cache.set({ ...validData, ckey: ckey + "A" });

    const all = await cache.getAll({ universe_name, project_name });
    expect(all.length).toEqual(2);
  });
  test("returns null when it can not find data", async () => {
    expect(
      await cache.get({ universe_name, project_name, ckey: "wrong" })
    ).toBeNull();
  });
});

describe("cache.clearCache()", () => {
  test("deletes all cache items", async () => {
    await cache.set(validData);
    await cache.set({ ...validData, ckey: ckey + "A" });

    const all = await cache.getAll({ universe_name, project_name });
    expect(all.length).toEqual(2);

    await cache.clearCache();

    const all2 = await cache.getAll({ universe_name, project_name });
    expect(all2.length).toEqual(0);
  });
});
