import {
  getMatchingFile,
  ICrashObjData,
  IMatchingGitFile,
} from "../src/lib/git_routines";
import db from "../src/lib/db";
import cache from "../src/lib/cache";
import { createAndCloneRepo, createAndCloneArchive } from "./mocks/git";
import * as fs from "fs";
import * as rimraf from "rimraf";
import { resolve } from "path";

const universeName = "testUniv";
const projectName = "testProj";
const crashObj = {} as ICrashObjData;

beforeAll(async () => {
  await db.migrate.latest();
  await cache.db.migrate.latest();
  rimraf.sync(resolve("test", "reposTmp"));
  // rimraf.sync(resolve("test", "corpusTmp"));
});

afterAll(async () => {
  await db.destroy();
  await cache.db.destroy();
});

describe("source code lookup", () => {
  test("getMatchingFile() returns null when nothing is found", async () => {
    const match = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "foo2", fullPath: "blah/foo2" },
      db,
    })) as IMatchingGitFile;
    expect(match).toBeNull();
  });
  test("getMatchingFile() returns a git file", async () => {
    // create a repo
    await createAndCloneRepo({
      db,
      universeName,
      projectName,
      script: [
        {
          msg: "my commit",
          files: [
            { filepath: "foo", txt: "this is foo" },
            { filepath: "blah/foo2", txt: "this is blah/foo2" },
          ],
        },
      ],
    });
    // create an archive
    // make sure we get the right file back
    const match = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "foo2", fullPath: "blah/foo2" },
      db,
    })) as IMatchingGitFile;
    expect(match).not.toBeNull();
    expect(match.filePath).toMatch(/foo2$/);
  });
  test("getMatchingFile() searches several repos", async () => {
    await db("repositories").delete();
    // create a repo
    await createAndCloneRepo({
      db,
      universeName,
      projectName,
      script: [
        {
          msg: "my commit",
          files: [
            { filepath: "foo", txt: "this is foo" },
            { filepath: "blah/foo2", txt: "this is blah/foo2" },
          ],
        },
      ],
    });
    // create a second repo
    await createAndCloneRepo({
      db,
      universeName,
      projectName,
      script: [
        {
          msg: "my other commit",
          files: [
            { filepath: "baz", txt: "this is baz" },
            { filepath: "buzz/quux", txt: "this is buzz/quux" },
          ],
        },
      ],
    });
    // make sure we get the right file back
    const match = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "foo2", fullPath: "blah/foo2" },
      db,
    })) as IMatchingGitFile;
    expect(match).not.toBeNull();
    expect(match.filePath).toMatch(/^blah\/foo2$/);
    // now try matching the other repo
    const match2 = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "baz", fullPath: "baz" },
      db,
    })) as IMatchingGitFile;
    expect(match2).not.toBeNull();
    expect(match2.filePath).toMatch(/^baz$/);
    // once more for good measure
    const match3 = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "quux", fullPath: "buzz/quux" },
      db,
    })) as IMatchingGitFile;
    expect(match3).not.toBeNull();
    expect(match3.filePath).toMatch(/^buzz\/quux$/);
  });
  test("getMatchingFile() searches archives", async () => {
    await db("repositories").delete();
    await createAndCloneRepo({
      db,
      universeName,
      projectName,
      script: [
        {
          msg: "my commit",
          files: [
            { filepath: "foo", txt: "this is foo" },
            { filepath: "blah/foo2", txt: "this is blah/foo2" },
          ],
        },
      ],
    });
    await createAndCloneArchive({
      name: "archie",
      universeName,
      projectName,
      script: [
        {
          msg: "archive commit",
          files: [
            { filepath: "hello", txt: "hello world" },
            { filepath: "yes/okay.js", txt: "console.log('hi')" },
          ],
        },
      ],
    });

    // find the repo file
    const match1 = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "foo2", fullPath: "blah/foo2" },
      db,
    })) as IMatchingGitFile;
    expect(match1).not.toBeNull();
    expect(match1.filePath).toMatch(/^blah\/foo2$/);

    // find the archive file
    const match2 = (await getMatchingFile({
      universeName,
      projectName,
      obj: { ...crashObj, fileName: "okay.js", fullPath: "yes/okay.js" },
      db,
    })) as IMatchingGitFile;
    expect(match2).not.toBeNull();
    expect(match2.filePath).toMatch(/yes\/okay.js$/);
  });
});
