"use strict";
const knexfile = require("../knexfile");
const knexfile_cachedb = require("../knexfile_cachedb");
const knex = require("knex");
const fs = require("fs");
const path = require("path");
const { spawnSync } = require("child_process");
const rimraf = require("rimraf");
import * as testUtils from "./test_utils";
import cache from "../src/lib/cache";
const cachedb = cache.db;
import { IDbCloneInfo } from "../src/models/repository";
import { withTimeout } from "../src/lib/util";
import * as gitRoutines from "../src/lib/git_routines";
import fetchQueue from "../src/lib/parallel_queue";

const {
  repoPath,
  readGitmodules,
  getSubmoduleCommits,
  getCommits,
  getCommit,
  gitLsTreeSubm,
  gitLsTree,
  gitLsTreeMatch,
  getSubmoduleInfoForDb,
  cloneSubmodulesRecursively,
  getCandidateRepos,
  cloneRepo,
  getUnclonedSubmodules,
  getUnclonedSubmodulesForProject,
  getSubmoduleAncestor,
  gitFetch,
  gitFetchExn,
  newGit,
  getRepo,
} = gitRoutines;
import { ICrashObjData } from "../src/lib/git_routines";
import DoNotUseList from "../src/lib/do_not_use_list";

const corpusDir = path.resolve("test", "corpus");

const universeName = "testUniv";
const projectName = "testProj";

function extractTestRepos(folder, tarfile = folder, repoName = "1") {
  // Extract repositories from tar file
  // repos must live in a tar file because it's not possible to add git repo
  // folders into another git repo.
  const testRepo = path.resolve(corpusDir, folder, repoName, ".git", "index");
  if (!fs.existsSync(testRepo)) {
    console.log(
      `${folder} test corpus does not exist, extracting:\n`,
      `tar xzvf test/corpus/${tarfile}.tar.gz -C test/corpus/`
    );
    spawnSync("tar", [
      "xzvf",
      path.resolve(corpusDir, `${tarfile}.tar.gz`),
      "-C",
      corpusDir,
    ]);
    if (!fs.existsSync(testRepo)) {
      process.on("exit", () =>
        console.log(
          `Unable to find ${folder} test corpus or corresponding ${tarfile}.tar.gz`
        )
      );
      process.exit(1);
    }
  }
}

beforeAll(async () => {
  try {
    // remove stale data
    rimraf.sync(path.resolve("test", "reposTmp"));
    fs.unlinkSync(knexfile.connection.filename);
  } catch (e) {}

  try {
    fs.unlinkSync(knexfile_cachedb.connection.filename);
  } catch (e) {}
  await cachedb.migrate.latest();
});

afterAll(async () => {
  await cachedb.destroy();
  try {
    fs.unlinkSync(knexfile_cachedb.connection.filename);
  } catch (e) {}
});

describe("file lookups work correctly", () => {
  const db = knex(knexfile);
  const folder = "repo_selection_test";
  beforeAll(async () => {
    extractTestRepos(folder);
    await db.migrate.latest();
    for (const repo of ["1", "2", "3"]) {
      const url = path.resolve(corpusDir, folder, repo);
      await cloneRepo(db, {
        name: repo,
        url: url,
        default_commit: "master",
        auth_type: "none",
        project_name: "testproj",
        universe_name: "testuniv",
        clone_status: undefined,
        user_name: undefined,
        password: undefined,
        ssh_public_key: undefined,
        ssh_private_key: undefined,
      });
    }
    expect(await db("repositories").select("id", "clone_status")).toEqual([
      { id: 1, clone_status: "complete" },
      { id: 2, clone_status: "complete" },
      { id: 3, clone_status: "complete" },
    ]);
  });
  afterAll(async () => {
    await db.destroy();
    rimraf.sync(path.resolve("test", "reposTmp"));
    fs.unlinkSync(knexfile.connection.filename);
  });
  test("gitLsTree lists all files", async () => {
    const repoId = 1;
    const repo = await getRepo(repoId);
    const correctFileList = ["1.txt", "A.txt"];
    const commit = await getCommit(repo, "8f49532");
    let lstree = await gitLsTree("testuniv", "testproj", repoId, repo, commit);
    expect(lstree).toMatchObject(correctFileList);
  });
  test("gitLsTree utilizes cache", async () => {
    const repoId = 1;
    const repo = await getRepo(repoId);
    const correctFileList = ["1.txt", "A.txt"];

    await testUtils.usesCache(async () => {
      const commit = await getCommit(repo, "8f49532");
      let lstree = await gitLsTree(
        "testuniv",
        "testproj",
        repoId,
        repo,
        commit
      );
      expect(lstree).toMatchObject(correctFileList);
    });
  });
  test("gitLsTreeMatch filters gitLsTree", async () => {
    const repoId = 1;
    const repo = await getRepo(repoId);
    const commit = await getCommit(repo, "8f49532");
    let lstree = await gitLsTreeMatch(
      "testuniv",
      "testproj",
      repoId,
      repo,
      commit,
      RegExp("A")
    );
    expect(lstree).toMatchObject(["A.txt"]);
  });
});

describe("repo selection rules work correctly", () => {
  const db = knex(knexfile);
  const folder = "repo_selection_test";
  beforeAll(async () => {
    extractTestRepos(folder);
    await db.migrate.latest();
    for (const repo of ["1", "2", "3"]) {
      const url = path.resolve(corpusDir, folder, repo);
      await cloneRepo(db, {
        name: repo,
        url: url,
        default_commit: "master",
        auth_type: "none",
        project_name: "testproj",
        universe_name: "testuniv",
        clone_status: undefined,
        user_name: undefined,
        password: undefined,
        ssh_public_key: undefined,
        ssh_private_key: undefined,
      });
    }
    expect(await db("repositories").select("id", "clone_status")).toEqual([
      { id: 1, clone_status: "complete" },
      { id: 2, clone_status: "complete" },
      { id: 3, clone_status: "complete" },
    ]);
  });
  test("getCandidateRepos returns all repos", async () => {
    // candidate repos should return all repos before we add rules
    expect(
      await getCandidateRepos(
        db,
        "testuniv",
        "testproj",
        {} as ICrashObjData,
        new DoNotUseList()
      )
    ).toMatchObject([
      { id: 1, clone_status: "complete" },
      { id: 2, clone_status: "complete" },
      { id: 3, clone_status: "complete" },
    ]);
  });
  test("getCandidateRepos respects MATCH rules", async () => {
    // add rule to prioritize repo 2 when function_name = foo
    // assert that repo 2 is placed at the top of the list
    await db("rules").insert({
      universe_name: "testuniv",
      project_name: "testproj",
      position: 0,
      matching_repository_id: 2,
      field_type: "function_name",
      field_name: "",
      regex: "foo",
      action: "match",
    });
    expect(
      await getCandidateRepos(
        db,
        "testuniv",
        "testproj",
        { functionName: "foo" } as ICrashObjData,
        new DoNotUseList()
      )
    ).toMatchObject([
      { id: 2, clone_status: "complete" },
      { id: 1, clone_status: "complete" },
      { id: 3, clone_status: "complete" },
    ]);
    await db("rules").delete();
  });
  test("getCandidateRepos respects SKIP rules", async () => {
    // add a rule that eliminates repo 2 when function_name = bar
    //   assert that repo1 isn't returned
    await db("rules").insert({
      universe_name: "testuniv",
      project_name: "testproj",
      position: 0,
      matching_repository_id: 2,
      field_type: "function_name",
      field_name: "",
      regex: "foo",
      action: "skip",
    });
    expect(
      await getCandidateRepos(
        db,
        "testuniv",
        "testproj",
        { functionName: "foo" } as ICrashObjData,
        new DoNotUseList()
      )
    ).toMatchObject([
      { id: 1, clone_status: "complete" },
      { id: 3, clone_status: "complete" },
    ]);
    await db("rules").delete();
  });
  afterAll(async () => {
    await db.destroy();
    rimraf.sync(path.resolve("test", "reposTmp"));
    fs.unlinkSync(knexfile.connection.filename);
  });
});

describe("submodules functions work correctly", () => {
  const db = knex(knexfile);
  beforeAll(async () => {
    extractTestRepos("submodule_test");
    await db.migrate.latest();
    for (const repo of ["1", "2", "3", "4", "7"]) {
      const url = path.resolve("test", "corpus", "submodule_test", repo);
      let submodule_parent_repo = null;
      if (repo === "2") {
        submodule_parent_repo = 1;
      }
      if (repo === "4") {
        submodule_parent_repo = 3;
      }
      if (repo === "7") {
        submodule_parent_repo = 4;
      }
      const [id] = await db("repositories").insert({
        id: repo,
        name: repo,
        url: url,
        default_commit: "master",
        auth_type: "none",
        project_name: "testproj",
        universe_name: "testuniv",
        clone_status: "complete",
        submodule_parent_repo,
      });
      const clonePath = path.resolve("test", "reposTmp", id.toString());
      if (fs.existsSync(clonePath)) {
        rimraf.sync(clonePath);
      }
      await newGit.cloneOpenExn(url, clonePath);
    }
  });

  test("readGitmodules works", async () => {
    const { id } = await db("repositories").first("id").where("name", "1");
    const modules = await readGitmodules(
      id,
      "2c0635d",
      universeName,
      projectName
    );
    expect(modules).toEqual({ 2: { path: "2", url: "../2", name: "2" } });
  });

  test("gitLsTreeSubm gets the correct submodule revisions", async () => {
    // this is the same as `git ls-tree -r -d 2c0635d`
    const { id } = await db("repositories").first("id").where("name", "1");
    const repo = await getRepo(id);
    const submCommits = await gitLsTreeSubm(1, repo, "2c0635d");
    expect(submCommits).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
    ]);
    const submCommits2 = await gitLsTreeSubm(1, repo, "f99fd80");
    expect(submCommits2).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
    ]);
    const submCommits3 = await gitLsTreeSubm(1, repo, "9fe8024");
    expect(submCommits3).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
      ["3", "a59791e4efa00303712b289a3f5859a5cec966dd"],
    ]);
    const submCommits4 = await gitLsTreeSubm(1, repo, "24799ba");
    expect(submCommits4).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
      ["3", "960296a6a91e8dd0c3553d68b1f592ee03682ed5"],
    ]);
    const submCommits5 = await gitLsTreeSubm(1, repo, "0df0fe2");
    expect(submCommits5).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
      ["3", "960296a6a91e8dd0c3553d68b1f592ee03682ed5"],
    ]);
    const submCommits6 = await gitLsTreeSubm(1, repo, "0525b58");
    expect(submCommits6).toEqual([
      ["2", "c773155aa9658eadfe2fe896758d6b5589077fe2"],
      ["3", "b1cb42268f0e6218870a1ed32c97cb9b445c9faa"],
    ]);
  });

  test("getSubmoduleCommits works", async () => {
    const repo = await db("repositories").first("*").where("name", "2");
    expect(
      await getSubmoduleCommits(repo, {}, db, universeName, projectName)
    ).toEqual(["c773155aa9658eadfe2fe896758d6b5589077fe2"]);
  });

  test("getCommits uses getSubmoduleCommits when appropriate", async () => {
    const repo = await db("repositories").first("*").where("name", "2");
    expect(await getCommits(repo, {}, db, universeName, projectName)).toEqual([
      "c773155aa9658eadfe2fe896758d6b5589077fe2",
    ]);
  });

  test("getUnclonedSubmodules discovers new, uncloned submodules", async () => {
    expect(
      await getUnclonedSubmodules(db, 3, "faf5d46", universeName, projectName)
    ).toMatchObject([
      { path: "5", url: "../5", name: "5" },
      { path: "6", url: "../6", name: "6" },
    ]);
  });

  test("lists new uncloned submodules", async () => {
    expect(
      await getUnclonedSubmodulesForProject(
        db,
        "testuniv",
        "testproj",
        {} as ICrashObjData
      )
    ).toMatchObject([
      { path: "5", url: "../5", name: "5" },
      { path: "6", url: "../6", name: "6" },
    ]);
  });

  test("recursively clones submodules", async () => {
    jest
      .spyOn(gitRoutines, "getSubmoduleInfoForDb")
      .mockImplementation(function (
        parentRepo: IDbCloneInfo,
        parentCommit: string,
        subName: string,
        subProps: any
      ) {
        let info = getSubmoduleInfoForDb(
          parentRepo,
          parentCommit,
          subName,
          subProps
        );
        info.url = path.resolve(
          "test",
          "corpus",
          "submodule_test",
          parentRepo.name,
          info.url
        );
        info.id = parseInt(info.name, 10);
        return info;
      });
    await cloneSubmodulesRecursively(
      db,
      3,
      "faf5d46",
      universeName,
      projectName
    );
    ["4", "5", "6", "7"].forEach((newSubmId) =>
      expect(async () => await getRepo(newSubmId)).not.toThrowError()
    );
  });
  test("recursively finds the root node of the repo/submodule tree", async () => {
    const parentId = (
      await getSubmoduleAncestor(
        db,
        await db("repositories").first("*").where("id", 6)
      )
    ).id;
    expect(parentId).toEqual(3);
  });

  afterAll(async () => {
    await db.destroy();
    rimraf.sync(path.resolve("test", "reposTmp"));
    fs.unlinkSync(knexfile.connection.filename);
  });
});

test("withTimeout() resolves faster promise first", async () => {
  async function twenty() {
    return new Promise((resolv, _rej) =>
      setTimeout(() => resolv("twenty"), 20)
    );
  }
  expect(await withTimeout(twenty(), 40)).toEqual("twenty");
  expect(await withTimeout(twenty(), 10)).toEqual("timeout");
});

describe("gitFetch", () => {
  const db = knex(knexfile);
  beforeAll(async () => {
    await db.migrate.latest();
  });
  afterAll(async () => {
    await db.destroy();
    // rimraf.sync(path.resolve("test", "reposTmp"));
    fs.unlinkSync(knexfile.connection.filename);
  });
  const testRepo = {
    name: "testRepo",
    url: "https://github.com/rashkov/dotfiles",
    default_commit: "master",
    auth_type: "none",
    project_name: "testproj",
    universe_name: "testuniv",
    clone_status: undefined,
    user_name: undefined,
    password: undefined,
    ssh_public_key: undefined,
    ssh_private_key: undefined,
  };
  test("gitFetchExn() throws proper errors", async () => {
    const id = await cloneRepo(db, testRepo);
    const repoDb = await db("repositories").first("*").where("id", id);
    const repo = await newGit.getRepo(id);
    // fetches without throwing
    await gitFetchExn(repo, repoDb);

    // fetches and throws
    const path = repoPath(id);
    await newGit.addOrUpdateRemote(path, "origin", testRepo.url + "corrupt");
    await expect(gitFetchExn(repo, repoDb)).rejects.toThrow(
      new RegExp("could not read username", "i")
    );
  });
  test("gitFetch() logs errors & last_fetched timestamp to db", async () => {
    const id = await cloneRepo(db, testRepo);
    const repoDb = await db("repositories").first("*").where("id", id);
    const repo = await newGit.getRepo(id);
    expect(repoDb.last_fetched).toBeNull();

    // logs last_fetched timestamp
    await gitFetch(repo, repoDb, db);
    let updatedRepoDb = await db("repositories").first("*").where("id", id);
    expect(updatedRepoDb.last_fetched).not.toBeNull();
    expect(updatedRepoDb.fetch_err).toBe("");

    // on fetch error, gitFetch():
    //   1. does not throw
    //   2. saves error to repo.fetch_err
    //   3. does not update the timestamp
    const goodUrl = repoDb.url;
    await newGit.addOrUpdateRemote(repoPath(id), "origin", goodUrl + "corrupt");
    // does not throw
    await expect(gitFetch(repo, repoDb, db)).resolves.not.toThrow();
    let updatedRepoDb2 = await db("repositories").first("*").where("id", id);
    // logs error to repo.fetch_err
    expect(updatedRepoDb2.fetch_err).toMatch(
      new RegExp("could not read username", "i")
    );
    // does not update timestamp
    expect(updatedRepoDb2.last_fetched).toBe(updatedRepoDb.last_fetched);

    // on subsequent fetch success, gitFetch():
    //   1. does not throw
    //   2. clears fetch_err
    //   3. updates timestamp
    await newGit.addOrUpdateRemote(repoPath(id), "origin", goodUrl);
    // does not throw
    await expect(gitFetch(repo, repoDb, db)).resolves.not.toThrow();
    let updatedRepoDb3 = await db("repositories").first("*").where("id", id);
    expect(updatedRepoDb3.last_fetched).not.toBeNull();
    expect(updatedRepoDb3.fetch_err).toBe("");
  });
});

describe("fetch queue", () => {
  const getTask = (ms: number) => (resolv: () => void, _rej: any) => {
    setTimeout(() => {
      resolv();
    }, ms);
  };
  test("it should accept items via push() & return promises", async () => {
    const q = new fetchQueue({ width: 2, freq: 50 });
    const promises = q.push([
      getTask(50),
      getTask(49),
      getTask(100),
      getTask(105),
    ]);
    expect(promises.length).toBe(4);
    expect(promises instanceof Array).toBe(true);
    expect(promises[0] instanceof Promise).toBe(true);
    q.destroy();
  });
  test("it should process those items in parallel", async () => {
    const q = new fetchQueue({ width: 2, freq: 5 });
    const start = new Date();
    const promises = q.push([
      getTask(50),
      getTask(10),
      getTask(10),
      getTask(49),
      getTask(100),
      getTask(105),
    ]);
    const resolved = [false, false, false, false, false, false];
    promises.forEach((p: Promise<any>, i: number) =>
      p.then(() => (resolved[i] = true))
    );
    await Promise.all(promises.slice(0, 3));
    // expect fourth fifth and sixth tasks to not be fulfilled
    expect(resolved).toEqual([true, true, true, false, false, false]);
    // expect elapsed time (new Date() - start) to be less than 60
    expect(start.valueOf() - new Date().valueOf()).toBeLessThan(60);
    q.destroy();
  });
  test("it should eventually process all items", async () => {
    const q = new fetchQueue({ width: 2, freq: 50 });
    const start = new Date();
    const promises = q.push([
      getTask(20),
      getTask(20),
      getTask(20),
      getTask(20),
    ]);
    const resolved = [false, false, false, false];
    promises.forEach(async (p: Promise<any>, i: number) => {
      await p;
      return (resolved[i] = true);
    });
    await Promise.all(promises);
    // expect all tasks to be complete
    expect(resolved).toEqual([true, true, true, true]);
    // expect elapsed time (new Date() - start) to be less than 50
    expect(start.valueOf() - new Date().valueOf()).toBeLessThan(50);
    q.destroy();
  });
  test("it should take more items via push()", async () => {
    const q = new fetchQueue({ width: 2, freq: 50 });
    const promises = q.push([getTask(20), getTask(20)]);
    const additional = q.push([getTask(20), getTask(20)]);
    const allPromises = promises.concat(additional);
    const resolved = [false, false, false, false];
    allPromises.forEach(async (p: Promise<any>, i: number) => {
      await p;
      return (resolved[i] = true);
    });
    await Promise.all(allPromises);
    // expect all tasks to be complete
    expect(resolved).toEqual([true, true, true, true]);
    q.destroy();
  });
});
